/*
    Digital circuits emulation library

    Copyright (c) 2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef DCE_74107_H_
#define DCE_74107_H_

/*
    Dual J-K flip-flop with clear.
*/
class D74107 : public DChip
{
public:
    D74107() : DChip(14) {
        //  1: 1J -> J input for flip-flop #1
        connectLine(  2, q1_not_ ); // NOT 1Q, output of flip-flop #1
        connectLine(  3, q1_ ); // 1Q, output of flip-flop #1
        //  4: 1K -> K input for flip-flop #1
        connectLine(  5, q2_ ); // 2Q, output of flip-flop #2
        connectLine(  6, q2_not_ ); // NOT 2Q, output of flip-flop #2
        connectLine(  7, DLine::Ground );
        //  8: 2J -> J input for flip-flop #2
        //  9: 2CLK -> clock for flip-flop #2
        // 10: 2CLR -> clear for flip-flop #2 (negated)
        // 11: 2K -> K input for flip-flop #2
        // 12: 1CLK -> clock for flip-flop #1
        // 13: 1CLR -> clear for flip-flop #1 (negated)
        connectLine( 14, DLine::Power );

        q1_ = 0;
        q1_not_ = 0;
        last_clk1_ = 0;
        clk1_edge_ = 0;
        q2_ = 0;
        q2_not_ = 0;
        last_clk2_ = 0;
        clk2_edge_ = 0;
    }

    void update() {
        // Flip-flop #1
        clk1_edge_ += last_clk1_ ^ line(12);

        if( ! line(13) ) {
            // Clear
            q1_ = 0;
            clk1_edge_ = 0;
        }
        else if( last_clk1_ && (clk1_edge_ >= 2) ) {
            q1_ = doFlipFlop( q1_, line( 1), line( 4) );
            clk1_edge_ = 0;
        }

        last_clk1_ = line(12);

        // Flip-flop #2
        clk2_edge_ += last_clk2_ ^ line( 9);

        if( ! line(10) ) {
            // Clear
            q2_ = 0;
            clk2_edge_ = 0;
        }
        else if( last_clk2_ && (clk2_edge_ >= 2) ) {
            q2_ = doFlipFlop( q2_, line( 8), line(11) );
            clk2_edge_ = 0;
        }

        last_clk2_ = line( 9);

        // Generate negative output too
        q1_not_ = 1 ^ q1_;
        q2_not_ = 1 ^ q2_;
    }

private:
    DBit doFlipFlop( DBit q, DBit j, DBit k ) {
        if( j ^ k ) { // J and K are different
            q = j;
        }
        else if( j ) { // J and K are both high
            q ^= 1;
        }

        return q;
    }

    DBit    q1_;
    DBit    q1_not_;
    DBit    last_clk1_;
    DBit    q2_;
    DBit    q2_not_;
    DBit    last_clk2_;
    unsigned    clk1_edge_;
    unsigned    clk2_edge_;
};

#endif // DCE_74107_H_
